;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - hierarchy-widget.ss                                            #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # Copyright (C) Peter Ivanyi, 2007                                               #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

; TODO:
; - set style: line connecting only to containers or line connecting to every element or no line

(module mreddesigner-hierarchy-widget mzscheme

  (require (lib "class.ss")  
           (lib "mred.ss" "mred")
  )

  (provide hierarchy-widget%)

; ------------------------------------------------------------------------------
; ------------------------------------------------------------------------------
;                                Hierarchy widget
; ------------------------------------------------------------------------------
; ------------------------------------------------------------------------------
 
;; We declare the class...
(define hierarchy-widget%
  ;; We derive it...
  (class canvas%

    ;; Overriden methods
    (override on-scroll on-size on-paint on-event on-char)

    ;; Public fields
    (init-field 
      (callback #f)
      (keyboard-callback #f)
    )
    ; check the callback function
    (unless (or (not callback) 
                (procedure-arity-includes? callback 2))
      (raise-type-error 'hierarchy-widget%
                        "procedure of arity 2"
                        callback)
    )
    ; check the keyboard callback function
    (unless (or (not keyboard-callback) 
                (procedure-arity-includes? keyboard-callback 3))
      (raise-type-error 'hierarchy-widget%
                        "procedure of arity 3"
                        keyboard-callback)
    )
    
    ; --------------------------------------------------------------------------
    ;                             Local variables, fields
    ; --------------------------------------------------------------------------

    ;; Fields
    ; this stores the id of the selected item
    (define item-selected #f)
    ; it stores the hierarchy in a list
    ; the first element in a list is the container
    ; a container can contain other containers, in this case the other container is a list,
    ; where the first element is the container itself
    (define item-list #f)
    ; internal properties, added after insertion
    ; - icon : a bitmap object
    (define item-table (make-hash-table 'equal))
    
    ;; Why using a bitmap and a drawing context as a class field instead
    ;; of directly declare it in the on-paint body? The matter is,
    ;; that if we did so a bitmap would be instantiated for each call of
    ;; the on-paint method. After testing it, even on a pretty powerful
    ;; pc, if the on-paint method is called too often, the program runs out
    ;; of resources and returns an error when instantiating a bitmap after
    ;; a short while.
    (define bitmap #f)
    (define dc #f)
    (define bitmap-width 0)
    (define bitmap-height 0)
    (define x-offset 0)
    (define y-offset 0)
    
    ;; Constants
    (define k-left-margin 21)
    (define k-top-margin 2)
    (define k-line-height 17)
    (define k-horizontal-scroll-step 3)
    
    ;; Those constants are used for drawing
    (define k-white         (make-object color% 255 255 255))
    (define k-black         (make-object color% 0 0 0))
    (define k-gray          (make-object color% 160 160 160))
    (define k-white-brush   (make-object brush% k-white 'solid))
    (define k-active-brush  (make-object brush% k-black 'hilite))
    ;; We use the default panel background (which depends on the platform, windowing system, etc.)
    (define k-gray-brush    (make-object brush% (get-panel-background) 'solid))
    ;; Opaque is needed here
    (define k-dkgray-brush  (make-object brush% (make-object color% 132 132 132) 'opaque))
    (define k-invisible-pen (make-object pen%   (make-object color% 0 0 0) 0 'transparent))
    (define k-edit-font     (make-object font%  12 "Courier" 'default 'normal 'normal #f 
                                                'unsmoothed #t))
    
    ;; To be platform compliant, we avoid smoothed text under windows, because highlighted smoothed
    ;; text under windows does not look really nice.
    (define k-normal-font  (if (eq? (system-type) 'windows)
                             (make-object font% 11 'system 'normal 'normal #f 'unsmoothed #t)
                             (make-object font% 11 'system 'normal 'normal #f 'smoothed #t)))
    
    (define k-italic-font  (if (eq? (system-type) 'windows)
                             (make-object font% 11 'system 'italic 'normal #f 'unsmoothed #t)
                             (make-object font% 11 'system 'italic 'normal #f 'smoothed #t)))
    
    (define k-bold-font    (if (eq? (system-type) 'windows)
                             (make-object font% 11 'system 'normal 'bold #f 'unsmoothed #t)
                             (make-object font% 11 'system 'normal 'bold #f 'smoothed #t)))
    
    ;; Loading the bitmaps now will allow us to reduce time needed to draw the widget.
    (define k-less-bitmap (make-object bitmap% (build-path "images" "less.png") 'png #f))
    (define k-plus-bitmap (make-object bitmap% (build-path "images" "plus.png") 'png #f))
    
    ; --------------------------------------------------------------------------
    ;                                Operations
    ; --------------------------------------------------------------------------
    
    ; for debugging
    (public get-table)
    (define (get-table)
      item-table
    )
    
    (public get-hierarchy)
    (define (get-hierarchy)
      item-list
    )
    
    (public get-item)
    (define (get-item id)
      (let*
        ((prop-table (hash-table-get item-table id
                                     (lambda ()
                                       (error "get-item: Cannot find id: " id))))
         (prop-list (hash-table-map
                      prop-table 
                      (lambda (key value) (list key value))
                    )
         )
        )
        prop-list
      )
    )
    
    (public get-selected)
    (define (get-selected)
      item-selected
    )
    
    (public set-selected)
    (define (set-selected id)
      (set! item-selected id)
      (on-paint)
      (if callback
        (callback 'select id)
      )
    )
    
    (public clear)
    (define (clear)
      (set! item-selected #f)
      (set! item-list #f)
      (set! item-table (make-hash-table 'equal))
      (update-bitmap)
    )
    
    ; --------------------------------------------------------------------------
    ;                              Adding items
    ; --------------------------------------------------------------------------
    
    (define update? #t)
    (public update-view!)
    (define (update-view! bool)
      (set! update? bool)
      (if bool
        (update-bitmap)
      )
    )
    
    
    ; append an id to the end of a container
    (define (append-id lst id parent)
      (cond
        ((null? lst) '())
        ((and (not (list? (car lst)))
              (equal? (car lst) parent))
         (append lst (list id))
        )
        ((list? (car lst))
         (cons (append-id (car lst) id parent)
               (append-id (cdr lst) id parent))
        )
        (else (cons (car lst)
                    (append-id (cdr lst) id parent))
        )
      )
    )
    
    ; add an item after parent id
    (define (add-after-id lst id parent)
      (cond
        ((null? lst) '())
        ((and (not (list? (car lst)))
              (equal? (car lst) parent))
         (append (list (car lst) id) (cdr lst))
        )
        ((list? (car lst))
         (cons (add-after-id (car lst) id parent)
               (add-after-id (cdr lst) id parent))
        )
        (else (cons (car lst)
                    (add-after-id (cdr lst) id parent))
        )
      )
    )
    
    ; add item after the item identified by parent id, 
    ; if parent is a container, add item after the container
    (define (add-after-item lst id parent)
      (cond
        ((null? lst) '())
        ((or (and (not (list? (car lst)))
                  (equal? (car lst) parent))
             (and (list? (car lst))
                  (equal? (caar lst) parent)))
         (append (list (car lst) id) (cdr lst))
        )
        ((list? (car lst))
         (cons (add-after-item (car lst) id parent)
               (add-after-item (cdr lst) id parent))
        )
        (else (cons (car lst)
                    (add-after-item (cdr lst) id parent))
        )
      )
    )
    
    ; add item before the item identified by parent id, 
    ; if parent is a container, add item before the container
    (define (add-before-item lst id before)
      (cond
        ((null? lst) '())
        ((or (and (not (list? (car lst)))
                  (equal? (car lst) before))
             (and (list? (car lst))
                  (equal? (caar lst) before)))
         (append (list id (car lst)) (cdr lst))
        )
        ((list? (car lst))
         (cons (add-before-item (car lst) id before)
               (add-before-item (cdr lst) id before))
        )
        (else (cons (car lst)
                    (add-before-item (cdr lst) id before))
        )
      )
    )
    
    (public item-append)
    (define (item-append parent id item icon container? overwrite?)
      ; type checks
      (if (not (string? id))
        (raise-type-error 'item-append "a string for id" id)
      )
      (if (not (list? item))
        (raise-type-error 'item-append "a list for an item" item)
      )
      (if (not (or (not icon)
                   (is-a? icon bitmap%)))
        (raise-type-error 'item-append "a bitmap for an icon" icon)
      )
      (if callback
        (callback 'before-item-append id)
      )

      ; convert association list to hash table
      (let
        ((table (make-hash-table 'equal)))
        (for-each
          (lambda (prop-list)
            (let
              ((name  (car prop-list))
               (value (cadr prop-list)))
              (hash-table-put! table name value)
            )
          )
          item
        )
        ; internal properties
        (hash-table-put! table 'collapsed #f)
        (hash-table-put! table 'icon icon)
        ; add to hash table
        (let 
          ((old-id (hash-table-get item-table id #f)))
          (if (or (and (not overwrite?) (not old-id))
                  overwrite?)
            (hash-table-put! item-table id table)
          )
        )
      )
      ; add to list
      (let
        ((id (if container? (list id) id)))
        (if item-list
          (if parent
            (set! item-list (append-id item-list id parent))
            (set! item-list (append-id item-list id (car item-list)))
          )
          (set! item-list id)
        )
      )
      (if update?
        (update-bitmap)
      )
      (if callback
        (callback 'after-item-append id)
      )
    )
    
    ; this is very similar to item-append !!!
    ; if after? is true it will add the widget after parent, 
    ; e.g. after a container and all of its widgets
    (public item-add)
    (define (item-add parent id item after? icon container? overwrite?)
      ; type checks
      (if (not (string? id))
        (raise-type-error 'item-add "a string for id" id)
      )
      (if (not (list? item))
        (raise-type-error 'item-add "a list for an item" item)
      )
      (if (not (or (not icon)
                   (is-a? icon bitmap%)))
        (raise-type-error 'item-add "a bitmap for an icon" icon)
      )
      (if callback
        (callback 'before-item-add id)
      )

      ; convert association list to hash table
      (let
        ((table (make-hash-table 'equal)))
        (for-each
          (lambda (prop-list)
            (let
              ((name  (car prop-list))
               (value (cadr prop-list)))
              (hash-table-put! table name value)
            )
          )
          item
        )
        ; internal properties
        (hash-table-put! table 'collapsed #f)
        (hash-table-put! table 'icon icon)
        ; add to hash table
        (let 
          ((old-id (hash-table-get item-table id #f)))
          (if (or (and (not overwrite?) (not old-id))
                  overwrite?)
            (hash-table-put! item-table id table)
          )
        )
      )
      ; add to list
      (let
        ((id (if container? (list id) id)))
        (if item-list
          (if parent
            (if after?
              (set! item-list (add-after-item item-list id parent))
              (set! item-list (add-after-id   item-list id parent))
            )
            (if after?
              (set! item-list (add-after-item item-list id (car item-list)))
              (set! item-list (add-after-id   item-list id (car item-list)))
            )
          )
          (set! item-list id)
        )
      )
      (if update?
        (update-bitmap)
      )
      (if callback
        (callback 'after-item-add id)
      )
    )
    
    ; --------------------------------------------------------------------------
    ;                           Deleting an item
    ; --------------------------------------------------------------------------
    
    ; deletes an id from the list 
    ; if the element to delete is the first in the list the whole list is deleted
    (define (del-item lst id)
      (cond
        ((null? lst) '())
        ((equal? (car lst) id) #f)
        (else
         (let
           ((rest (map (lambda (x) 
                         (cond
                           ((list? x)
                            (let ((res (del-item x id)))
                              (if res (list res) '())))
                           ((equal? x id) '())
                           (else (list x))))
                       lst)))
           (apply append rest)))))
    
    (public item-delete)
    (define (item-delete id)
      (if item-list
        (begin
          (if callback
            (callback 'before-item-delete id)
          )
          (set! item-list (del-item item-list id))
          (hash-table-remove! item-table id)
          (set! item-selected #f)
          (update-bitmap)
;          (on-paint)
          (if callback
            (callback 'after-item-delete id)
          )
        )
      )
    )
    
    ; --------------------------------------------------------------------------
    ;                             Property handling
    ; --------------------------------------------------------------------------
    
    ; returns a property of an item
    (public item-property-get)
    (define (item-property-get id prop allow-empty?)
      (let* 
        ((prop-table (hash-table-get item-table id
                                     (lambda ()
                                       (error "item-property-get: Cannot find id for property: "
                                              (list id prop)))))
         (value  (hash-table-get prop-table prop
                                 (lambda ()
                                   (if allow-empty?
                                     #f
                                     (error "item-property-get: Cannot find id for property: " 
                                            (list id prop))))))
        )
        value
      )
    )
    
    ; sets a property for an item
    (public item-property-set!)
    (define (item-property-set! id prop value notify?)
      (if (and notify? callback)
        (callback 'before-item-property-set (list id prop value))
      )
      (let* 
        ((prop-table (hash-table-get item-table id
                                     (lambda ()
                                       (error (string-append "item-property-set!: "
                                                             "Cannot find id for property and value: ")
                                              (list id prop value)))))
        )
        (hash-table-put! prop-table prop value)
      )
      (if update?
        (update-bitmap)
      )
      (if (and notify? callback)
        (callback 'after-item-property-set (list id prop value))
      )
    )
    
    
    ; helper functions
    (define (children? id)
      (item-property-get id 'children #t)
    )
    
    (define (collapsed? id)
      (item-property-get id 'collapsed #f)
    )
    
    ; --------------------------------------------------------------------------
    ;                       Swapping up or down
    ; --------------------------------------------------------------------------
    
    (define (can-swap-up? lst id)
      (define (can-swap-up-aux? lst id prev)
        (cond
          ((null? lst) #f)
          ((list? (car lst))
           (if (equal? (caar lst) id)
             (if prev prev #f)
             (let ((prev (can-swap-up-aux? (cdar lst) id #f)))
               (if prev
                 prev
                 (can-swap-up-aux? (cdr lst) id (car lst))
               )
             )
           )
          )
          ((equal? (car lst) id)
           (if prev prev #f)
          )
          (else
           (can-swap-up-aux? (cdr lst) id (car lst))
          )
        )
      )
      (let ((prev (can-swap-up-aux? (cdr lst) id #f)))
        (if (list? prev)
          (car prev)
          prev
        )
      )
    )
    
    (public item-can-swap-up?)
    (define (item-can-swap-up? id)
      (can-swap-up? item-list id)
    )
    
    (define (can-swap-down? lst id)
      (define (can-swap-down-aux? lst id)
        (cond
          ((null? lst) #f)
          ((list? (car lst))
           (if (equal? (caar lst) id)
             (if (not (null? (cdr lst))) (cadr lst) #f)
             (let ((next (can-swap-down-aux? (cdar lst) id)))
               (if next
                 next
                 (can-swap-down-aux? (cdr lst) id)
               )
             )
           )
          )
          ((equal? (car lst) id)
           (if (not (null? (cdr lst))) (cadr lst) #f)
          )
          (else
           (can-swap-down-aux? (cdr lst) id)
          )
        )
      )
      (let ((next (can-swap-down-aux? (cdr lst) id)))
        (if (list? next)
          (car next)
          next
        )
      )
    )
    
    (public item-can-swap-down?)
    (define (item-can-swap-down? id)
      (can-swap-down? item-list id)
    )
    
    (define (item-swap lst item)
      (cond
        ((null? lst) '())
        ((list? (car lst))
         (if (equal? (caar lst) item)
           (append (list (cadr lst) (car lst)) (cddr lst))
           (cons (item-swap (car lst) item) (item-swap (cdr lst) item))
         )
        )
        ((equal? (car lst) item)
         (append (list (cadr lst) (car lst)) (cddr lst))
        )
        (else
         (cons (car lst) (item-swap (cdr lst) item))
        )
      )
    )
    
    (public item-swap-up)
    (define (item-swap-up item)
      (let ((prev (can-swap-up? item-list item)))
        (if prev
          (begin
            (if callback
              (callback 'before-swap-up item)
            )
            (set! item-list (item-swap item-list prev))
            (update-bitmap)
;            (on-paint)
            (if callback
              (callback 'after-swap-up item)
            )
          )
        )
      )
    )
    
    (public item-swap-down)
    (define (item-swap-down item)
      (let ((next (can-swap-down? item-list item)))
        (if next
          (begin
            (if callback
              (callback 'before-swap-down item)
            )
            (set! item-list (item-swap item-list item))
            (update-bitmap)
;            (on-paint)
            (if callback
              (callback 'after-swap-down item)
            )
          )
        )
      )
    )
    
    ; ---------------------------------------------------------------
    ;                         Move one up
    ; ---------------------------------------------------------------
    
    ; returns #f or a five element list:
    ; element - pparent - parent - prev - first?
    (define (can-move-up? lst id condition)
      (let ((result #f))
        (define (can-move-up-aux? lst id pparent parent prev first?)
          (cond
            ((null? lst) #f)
            ((or (equal? (car lst) id)
                 (and (list? (car lst)) (equal? (caar lst) id)))
             (set! result (list (car lst) pparent parent prev first?))
             #t
            )
            (else
              (let ((res #f))
                (if (list? (car lst))
                  (set! res (can-move-up-aux? (cdar lst) id parent (caar lst) prev #t))
                )
                (if res
                  res
                  (can-move-up-aux? (cdr lst) id pparent parent (car lst) #f)
                )
              )
            )
          )
        )
        (can-move-up-aux? (list lst) id #f #f #f #f)
        (let*
          ((obj     (list-ref result 0))
           (pparent (list-ref result 1))
           (parent  (list-ref result 2))
           (prev    (list-ref result 3))
           (first?  (list-ref result 4)) )
          ; exclude obvious cases:
          (if (or (and (not pparent) (not parent) (not prev) (not first?)) ; the root
                  (and (equal? parent (car lst)) first?)) ; the first element in the root
            #f
            ; user condition
            (if (procedure? condition)
              (condition lst id result)
              result
            )
          )
        )
      )
    )
    
    (public item-can-move-up?)
    (define (item-can-move-up? item condition)
      (can-move-up? item-list item condition)
    )
    
    (define (move-up-one lst id move?)
      (let*
        ((result move?))
        (if result
          (let 
            ((obj     (list-ref result 0))
             (pparent (list-ref result 1))
             (parent  (list-ref result 2))
             (prev    (list-ref result 3))
             (first?  (list-ref result 4))
             ; delete the element
             (lst     (del-item lst id)))
            (cond
              ; add current to parent of the parent of the current, 
              ; just before the parent
              ((and first? pparent)
               (add-before-item lst obj parent)
              )
              ; the previous is a container so add to it
              ((and prev (list? prev))
               (let ((last (list-ref prev (- (length prev) 1))))
                 (if (list? last)
                   (add-after-item lst obj (car last))
                   (add-after-id   lst obj last))
               )
              )
              ; current was not the first in the list, therefore there is a prev
              ; and add current before prev
              (else
                (add-before-item lst obj prev)
              )
            )
          )
          lst
        )
      )
    )
    
    (public item-move-up)
    (define (item-move-up item condition)
      (let ((move? (can-move-up? item-list item condition)))
        (if move?
          (begin
            (if callback
              (callback 'before-move-up item)
            )
            (set! item-list (move-up-one item-list item move?))
            (update-bitmap)
;            (on-paint)
            (if callback
              (callback 'after-move-up item)
            )
          )
        )
      )
    )
    
    ; ---------------------------------------------------------------
    ;                         Move one down
    ; ---------------------------------------------------------------
    
    ; returns #f or a five element list:
    ; element - pparent - parent - next - last?
    (define (can-move-down? lst id condition)
      (let ((result #f))
        (define (can-move-down-aux? lst id pparent parent next last?)
          (cond
            ((null? lst) #f)
            ((or (equal? (car lst) id)
                 (and (list? (car lst)) (equal? (caar lst) id)))
             (let ((last? (null? (cdr lst))))
               (set! result (list (car lst) pparent parent 
                                  (if last? next (cadr lst)) 
                                  last?))
             )
             #t
            )
            (else
              (let ((res #f))
                (if (list? (car lst))
                  (set! res (can-move-down-aux? (cdar lst) id parent (caar lst)
                                                (if (null? (cdr lst))
                                                  next
                                                  (cadr lst)
                                                )
                                                #f))
                )
                (if res
                  res
                  (can-move-down-aux? (cdr lst) id pparent parent next #f)
                )
              )
            )
          )
        )
        (can-move-down-aux? (list lst) id #f #f #f #f)
        (let*
          ((obj     (list-ref result 0))
           (pparent (list-ref result 1))
           (parent  (list-ref result 2))
           (next    (list-ref result 3))
           (last?   (list-ref result 4)) )
          ; exclude obvious cases:
          (if (or (and (not pparent) (not parent) (not next) last?) ; the root
                  (and (equal? parent (car lst)) last?)) ; the last element in the root
            #f
            ; user condition
            (if (procedure? condition)
              (condition lst id result)
              result
            )
          )
        )
      )
    )
    
    (public item-can-move-down?)
    (define (item-can-move-down? item condition)
      (can-move-down? item-list item condition)
    )
    
    (define (move-down-one lst id move?)
      (let*
        ((result move?))
        (if result
          (let 
            ((obj     (list-ref result 0))
             (pparent (list-ref result 1))
             (parent  (list-ref result 2))
             (next    (list-ref result 3))
             (last?   (list-ref result 4))
             ; delete the element
             (lst     (del-item lst id)))
            (cond
              ; add current to parent of the parent of the current, 
              ; just after the parent
              ((and last? pparent)
               (add-after-item lst obj parent)
              )
              ; the next is a container so add to it
              ((and next (list? next))
               (add-after-id lst obj (car next))
              )
              ; current was not the last in the list, therefore there is a next
              ; and add current after next
              (else
               (add-after-id lst obj next)
              )
            )
          )
          lst
        )
      )
    )
    
    (public item-move-down)
    (define (item-move-down item condition)
      (let ((move? (can-move-down? item-list item condition)))
        (if move?
          (begin
            (if callback
              (callback 'before-move-down item)
            )
            (set! item-list (move-down-one item-list item move?))
            (update-bitmap)
;            (on-paint)
            (if callback
              (callback 'after-move-down item)
            )
          )
        )
      )
    )
    
    ; --------------------------------------------------------------------------
    ;                      Update canvas and scrollbars
    ; --------------------------------------------------------------------------
    
    ; width of all text of an item in pixel
    (define (item-text-width-get id)
      (let-values 
        (((label-width label-height label-distance label-extra)
          (get-label-extent id)
         )
         ;; Ok, we found the label width - we need to check the id width, now.
         ((id-width id-height id-distance id-extra)
          (get-id-extent id)
         ))
         (+ label-width id-width)
      )
    )
    
    ; the width of the line of an item in pixel
    (define (item-line-width-get id offset)
      (+ (* 20 offset) k-left-margin 20 (item-text-width-get id) 10)
    )
    
    ;; This function returns the "real" width of the widget, i.e.
    ;; the total width needed to display the whole widget's contents.
    (define (get-real-width)
      (define (get-real-width-aux data-list offset)
        (cond
          ((or (not data-list) (null? data-list)) 0)
          ((list? data-list)
           (if (collapsed? (car data-list))
             (item-line-width-get (car data-list) offset)
             (let
               ((w (item-line-width-get (car data-list) offset)))
               (for-each
                 (lambda (data) 
                   (set! w (max w (get-real-width-aux data (add1 offset))))
                 )
                 (cdr data-list)
               )
               w
             )
           )
          )
          (else (item-line-width-get data-list offset))
        )
      )
      
      ;; And then, we call the function we just declared...
      (inexact->exact (get-real-width-aux item-list 0)))
    
    ;; This function returns the "real" height of the widget, i.e.
    ;; the total height needed to display the whole widget's contents.
    (define (get-real-height)
      ;; We add 1 to ensure the height is != 0...
      (add1 (* k-line-height (children-num-r-nc item-list))))
        
    ;; Returns total number of children
    (define (children-num-r data-list)
      (cond
        ((not data-list) 0)
        ((list? data-list)
         (if (null? data-list)
           0
           (+ 1 (apply + (map (lambda (e) (if (list? e) (children-num-r e) 1))
                              (cdr data-list))))
         )
        )
        (else 1)
      )
    )
        
    ;; Returns total number of non-collapsed children
    (define (children-num-r-nc data-list)
      (cond
        ((not data-list) 0)
        ((and (list? data-list)
              (null? data-list))
         0
        )
        ((and (list? data-list)
              (not (collapsed? (car data-list))))
           (+ 1 (apply + (map (lambda (e) (if (list? e) (children-num-r-nc e) 1))
                              (cdr data-list))))
        )
        (else 1)
      )
    )
    
    ;; This function updates the bitmap in which drawings are performed before
    ;; appearing on the screen.
    (define (update-bitmap)
      (update-scroll-bars)
      ;; We will need the virtual size...
      (let-values (((width height) (send this get-virtual-size)))
        ;; We create a bitmap, with the *real* width and height (at least, if they are
        ;; greater than the virtual width and height...), 
        ;; in order to be able to scroll very quickly...
        (let*
          ((w (max (get-real-width) width 1))
           (h (max (get-real-height) height 1)))
          (if (or (not bitmap)
                  (> w bitmap-width)
                  (> h bitmap-height))
            (let*
              ((n (round (max (/ w 1000) (/ h 1000))))
               (size (* (+ n 1) 1000)))
              (set! bitmap (make-object bitmap% size size #f))
              ;; And then, we update the associated bitmap-dc...
              (set! dc (instantiate bitmap-dc% (bitmap)))
              (set! bitmap-width size)
              (set! bitmap-height size)
            )
          )
        )
      )
      (on-paint)
    )

    ;; This function updates the scroll bars properties...
    (define (update-scroll-bars)
      (let
        ((horizontal? #f) (vertical? #f))
        ;; hide both of the scroll bars
        (send this show-scrollbars #f #f)
        ;; We need to get the virtual size...
        (let-values (((width height) (send this get-virtual-size)))
          ;; Then, we update the scroll range for both the horizontal and 
          ;; the vertical scroll bars. 
          ;; The right range is the full real height (width, respectively) minus the
          ;; virtual height (resp. width).
          ;; vertical scrollbar
          (send this set-scroll-range 'vertical (max 0 (- (get-real-height) height)))
          ;; If the window can contain the whole height, no y offset is needed.
          (if (= 0 (send this get-scroll-range 'vertical)) 
            (set! y-offset 0)
            (begin
              (send this show-scrollbars #f #t)
              (set! vertical? #t)
            )
          )
        )
        (let-values (((width height) (send this get-virtual-size)))
          ;; horizontal scrollbar
          (send this set-scroll-range 'horizontal (max 0 (- (get-real-width) width)))
          ;; If the window can contain the whole width, no x offset is needed.
          (if (= 0 (send this get-scroll-range 'horizontal))
            (set! x-offset 0) 
            (begin
              (send this show-scrollbars #t vertical?)
              (set! horizontal? #t)
            )
          )
        )
        (let-values (((width height) (send this get-virtual-size)))
          ;; check vertical scrolling again
          (send this set-scroll-range 'vertical (max 0 (- (get-real-height) height)))
          ;; If the window can contain the whole height, no y offset is needed.
          (if (= 0 (send this get-scroll-range 'vertical))
            (set! y-offset 0)
            (send this show-scrollbars horizontal? vertical?)
          )
          ;; This is the number of scroll steps - 
          ;; vertically we want to scroll not exactly one page, 
          ;; but one page minus one line (as it is always implemented).
          (send this set-scroll-page 'horizontal (max 1 width))
          (send this set-scroll-page 'vertical (max 1 (- height k-line-height)))
        )
      )
    )
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-scroll                                 |   
    ;; +----------------------------------------------------------------------+
    (define (on-scroll scroll-event)
      ;; What we are going to do depends on the event direction...
      (case (send scroll-event get-direction)
        ;; Is it vertical?
        ((vertical) 
         ;; Then it depends on the event type...
         (let ((type (send scroll-event get-event-type)))
           ;;
           ;; Notice : I am forced to use eq?, because 
           ;; I have noticed a bug under MacOS X : if the event is a thumb event,
           ;; DrScheme crashes if we try to evaluate the result given by get-event-type! 
           ;; It looks to return an invalid pointer, at a system adress, maybe 0. 
           ;; Therefore I just compare pointers.
           ;;
           ;; If the event type is line down...
           (cond ((eq? type 'line-down)
                  ;; ...then we need the virtual size...
                  (let-values (((width height) (send this get-virtual-size)))
                    ;; ...update the scroll-bars positions 
                    ;; (to force scrolling more than 1 step [1 line, actually])...
                    (send this set-scroll-pos 'vertical 
                          (min (+ (send this get-scroll-pos 'vertical) k-line-height -1) 
                               (send this get-scroll-range 'vertical)))
                    ;; ...and the vertical offset.
                    (set! y-offset (min (+ y-offset k-line-height) 
                                        (send this get-scroll-range 'vertical)))))
                 
                 ;; If the event type is line-up...
                 ((eq? type 'line-up)
                  ;; ...then we update the scroll-bars positions 
                  ;; (to force scrolling more than 1 step...)...
                  (send this set-scroll-pos 'vertical 
                        (max (- (send this get-scroll-pos 'vertical) k-line-height -1) 0))
                  ;; ...and the vertical offset.
                  (set! y-offset (max (- y-offset k-line-height) 0)))
                 
                 ;; Otherwise, we do not have to update the scroll-bars positions, 
                 ;; just to update the vertical offset...
                 (else (set! y-offset (send scroll-event get-position))))))
        
        ;; Is it horizontal?
        ((horizontal) 
         ;; Then it depends on the event type...
         (let ((type (send scroll-event get-event-type)))
           ;; If the event type is line down...
           (cond ((eq? type 'line-down)
                  ;; ...then we need the virtual size...
                  (let-values (((width height) (send this get-virtual-size)))
                    ;; ...update the scroll-bars positions 
                    ;; (to force scrolling more than 1 step... [1 "horizontal scroll step", actually])...
                    (send this set-scroll-pos 'horizontal 
                          (min (+ (send this get-scroll-pos 'horizontal) k-horizontal-scroll-step -1) 
                                  (send this get-scroll-range 'horizontal)))
                    ;; ...and the horizontal offset.
                    (set! x-offset (min (+ x-offset k-horizontal-scroll-step) 
                                        (send this get-scroll-range 'horizontal)))))
                 
                 ;; If the event type is line-up...
                 ((eq? type 'line-up)
                  ;; ...then we update the scroll-bars positions 
                  ;; (to force scrolling more than 1 step...)...
                  (send this set-scroll-pos 'horizontal 
                        (max (- (send this get-scroll-pos 'horizontal) k-horizontal-scroll-step -1) 0))
                  ;; ...and the horizontal offset.
                  (set! x-offset (max (- x-offset k-horizontal-scroll-step) 0)))
                 
                 ;; Otherwise, we do not have to update the scroll-bars positions, 
                 ;; just to update the horizontal offset...
                 (else (set! x-offset (send scroll-event get-position)))))))
      
      ;; And then, we directly draw the bitmap, which has already been drawed during the last
      ;; call of (on-paint), at the right position - which is *much* faster than calling on-paint
      ;; directly and allows a smooth scrolling.
      (send (send this get-dc) draw-bitmap bitmap (- x-offset) (- y-offset)))
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-size                                   |   
    ;; +----------------------------------------------------------------------+
    (define (on-size width height)
      ;; We just update the bitmap...
      (update-bitmap)
    )
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-paint                                  |
    ;; +----------------------------------------------------------------------+
    
    (define (font-type-set id)
      (cond
        ; FIXME: this is not general enough
        ((memq (item-property-get id 'type #f) '(frame project)) k-bold-font)
        ;((container? id) k-italic-font)
        (else k-normal-font))
    )
    
    ;; it returns a string
    (define (get-label-text id)
      (let
        ((label (item-property-get id 'label #t)))
        (if label
          (if (not (string=? label ""))
            label
            "(no label)")
          (string-append "<" id ">")
        )
      )
    )
    
    ;; Determine the extent of the label text
    ;; it returns four values
    (define (get-label-extent id)
      (send dc get-text-extent 
            ;; The label text...
            (get-label-text id)
            ;; And its font...
            (font-type-set id))
    )
    
    ;; it returns a string
    (define (get-id-text id)
      (string-append " [" id "]")
    )
    
    ;; Determine the extent of the ID text
    ;; it returns four values
    (define (get-id-extent data-list)
      (send dc get-text-extent 
            (get-id-text data-list)
            ;; The id is drawn in an italic font...
            k-italic-font)
    )
    
    (define (item-draw-one item-id incr offset container?)
      (let
        ((item-type (item-property-get item-id 'type #f)))
        
        ;; Text foreground is black...
        (send dc set-text-foreground k-black)
                
        ;; Set the right font...
        (send dc set-font (font-type-set item-id))
                
        ;; The root name is on a gray background.
        (if (zero? offset)
          (begin
            (send dc set-brush k-gray-brush)
            (send dc draw-rectangle 0 
                  (* incr k-line-height) (send bitmap get-width) k-line-height)))
        
        ;; Let's draw the links betweeb the +/- boxes.
        ;; First we choose the color...
        (send dc set-brush k-dkgray-brush)
        
        ;; We draw the right icon...
        (if (item-property-get item-id 'icon #t)
          (send dc draw-bitmap
                (item-property-get item-id 'icon #t)
                (+ (* 20 offset) k-left-margin)
                (+ (* k-line-height incr) 1) 
                'solid 
                (make-object color% "black")
                #f))
                
        ;; Draw the +/- icons
        (if container?
          ;; If the current element is collapsed...
          (if (not (collapsed? item-id))
            ;; ...we draw a +...
            (send dc draw-bitmap 
                  k-less-bitmap
                  (+ (* 20 offset) 5)
                  (+ (* k-line-height incr) 4) 
                  'solid (make-object color% "white") #f)
            ;; ...we draw a - otherwise.
            (send dc draw-bitmap 
                  k-plus-bitmap
                  (+ (* 20 offset) 5)
                  (+ (* k-line-height incr) 4) 
                  'solid (make-object color% "white") #f)))
                
        (send dc set-text-foreground k-black)

        ;; Display the caption...
        (send dc draw-text
              (get-label-text item-id)
              (+ (* 20 offset) k-left-margin 20) 
              (+ (* k-line-height incr) k-top-margin))
                
        (let-values (((width height distance extra) (get-label-extent item-id)))
          (send dc set-text-foreground k-gray)
                  
          ;; Display the id...
          (send dc set-font k-italic-font)
          (send dc draw-text
                (get-id-text item-id)
                (+ (* 20 offset) k-left-margin 20 width)
                (+ (* k-line-height incr) k-top-margin))
                  
          ;; Is the current element selected?
          (if (equal? item-id item-selected)
            ;; Then draw the selection!
            (begin
              (send dc set-brush k-active-brush)
              (send dc draw-rectangle 
                    (+ (* 20 offset) k-left-margin 19) 
                    (+ (* k-line-height incr) distance -1)
                    (+ width 2)
                    (- k-line-height 2)))))
      )
    )
    
    
    (define (on-paint)
      (let* ((width  (send this get-width))
             (height (send this get-height))
             (incr   0)
            )
        
        ;; function to draw all lines recursively
        (define (draw-data-list data-list offset)
          (cond
            ((not data-list) 0)
            ((null? data-list) 0)
            ((list? data-list)
             (item-draw-one (car data-list) incr offset #t)
             (let ((start incr))
               (set! incr (add1 incr))
               (if (not (collapsed? (car data-list)))
                 (for-each
                   (lambda (x)
                     ; if container, then draw connecting line
                     ;(if (list? x)
                       (begin
                         ; horizontal line
                         (send dc draw-rectangle 
                               (+ (* 20 offset) 10) (+ (* (+ 0 incr) k-line-height) 8)
                               24 1)
                         ; vertical line
                         (send dc draw-rectangle 
                               (+ (* 20 offset) 9)  (+ (* (- (+ start 1) 0.5) k-line-height) 4)
                               1 (+ (* (- incr start 0.5) k-line-height) 5))
                       )
                     ;)
                     (draw-data-list x (add1 offset))
                   )
                   (cdr data-list)
                 )
               )
             )
            )
            (else
             (item-draw-one data-list incr offset #f)
             (set! incr (+ incr 1))
            )
          )
        )
        
        (send dc set-pen k-invisible-pen) ;; Invisible pen
        (send dc set-brush k-white-brush) ;; White brush

        ;; We clear the bitmap...
        (send dc clear)

        ;; ...and then parse the list
        (draw-data-list item-list 0)
        
        ;; We drawed the whole output into a bitmap in order to avoid flickering,
        ;; therefore we have to draw the bitmap.
        (if (send bitmap ok?)
          (send (send this get-dc) draw-bitmap bitmap (- x-offset) (- y-offset))
        )
      )
    )
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-event                                  |   
    ;; +----------------------------------------------------------------------+
    (define (action-select id)
      (set! item-selected id)
    )
    
    (define (action-collapse id)
      (let
        ((current (item-property-get id 'collapsed #f)))
        (item-property-set! id 'collapsed (not current) #f)
      )
    )
    
    (define (item-action-get id x y incr offset)
      (let-values 
        (((width height distance extra) (get-label-extent id)))
        
        (cond
          ;; Is the event a field area?
          ((and (>= x (+ (* 20 offset) k-left-margin ))
                (<= x (+ (* 20 offset) k-left-margin width 20))
                (>= y (+ (* k-line-height incr) distance -1))
                (<= y (+ (* k-line-height incr) distance 1 height)))
           (action-select id)
           (cons 'select id)
          )
          ;;Is the event in a +/- area?
          ((and (>= x (+ (* 20 offset) k-left-margin -16))
                (<= x (+ (* 20 offset) k-left-margin -7))
                (>= y (+ (* k-line-height incr) 4))
                (<= y (+ (* k-line-height incr) 13)))
           (action-select id)
           (action-collapse id)
           (cons 'collapse id)
          )
          (else #f)
        )
      )
    )

    (define (on-event event)
      ;; The widget gets the focus
      (send this focus)
      
      ;; On-event body starts here
      (let ((x        (+ (send event get-x) x-offset))
            (y        (+ (send event get-y) y-offset))
            (action   #f)
            (selected #f)
            (incr     0)
           )
        
        ; this function iterates through all non-collapsed objects
        ; if it finds a matching objects returns the action
        (define (check-data-list data-list offset)
          (cond
            ((not data-list) #f)
            ((null? data-list) #f)
            ((list? data-list)
             (if (not (collapsed? (car data-list)))
               (let ((action (check-data-list (car data-list) offset)))
                 (if action
                   action
                   (let loop ((data-list (cdr data-list)))
                     (if (null? data-list)
                       #f
                       (let ((action (check-data-list (car data-list) (add1 offset))))
                         (if action
                           action
                           (if (null? (cdr data-list)) #f (loop (cdr data-list)))
                         )
                       )
                     )
                   )
                 )
               )
               (check-data-list (car data-list) offset)
             )
            )
            (else
             (let ((action (item-action-get data-list x y incr offset)))
               (set! incr (+ incr 1))
               action
             )
            )
          )
        )
        
        (case (send event get-event-type)
          ;; Mouse-down event with the left mouse button
          ((left-down)
           (let
             ((action (check-data-list item-list 0)))
             (on-paint)
             
             (if (and action callback)
               (callback (car action) (cdr action))
             )
             
             )))))
    
    ;; +----------------------------------------------------------------------+
    ;; | Function properties-widget::on-char                                  |   
    ;; +----------------------------------------------------------------------+
    (define (on-char event)
      (let
        ((key (send event get-key-code)))
        (case key
          ;; Wheel up
          ((wheel-up)
           (on-scroll (new scroll-event% (event-type 'line-up) (direction 'vertical))))
          ;; Wheel down
          ((wheel-down)
           (on-scroll (new scroll-event% (event-type 'line-down) (direction 'vertical))))
          (else
           (if keyboard-callback
             (keyboard-callback key
                                (send event get-shift-down)
                                (send event get-control-down)
             )
           )
          )
        )
      )
    )
            

    ;; Call the initialization function from the parent class...
    (super-new (style '(vscroll hscroll border)))
    ;; ...ok, now we can call 'this'.
    
    ;; We have to initialize the scrollbars
    (send this init-manual-scrollbars #f #f 1 1 0 0)
    
    ;; We then need to call on-size so that the bitmap and
    ;; drawing context are initialised
    (on-size (send (send this get-parent) get-width) (send (send this get-parent) get-height))))

)


