;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - preview.ss                                                     #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # Copyright (C) Peter Ivanyi, 2007                                               #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

;; This file contains general functions used in previewing.

(module mreddesigner-preview mzscheme

  (require (lib "class.ss")  
           (lib "mred.ss" "mred")
           (file "mreddesigner-widgets.ss")
           (file "mreddesigner-misc.ss")
           (file "mreddesigner-default-values.ss")
  )
  
  ; when the user clicks on a widget in the preview window
  ; this general callback function is called for all kinds of widgets
  (define (preview-clicked-callback id)
    (let
      ((selected (send hierarchy-widget get-selected)))
      (if (not (equal? id selected))
        (send hierarchy-widget set-selected id)
      )
    )
  )
  
  
  (provide make-widget-preview)
  (define (make-widget-preview parent-id props)
    (let* 
       ; remove internal properties if any
      ((item-props (assoc-remove props '(collapsed icon pointer)))
       ; determine the type
       (type (cadr (assoc 'type props)))
       ; generate an id symbol
       (id-sym (gensym (string->symbol (string-append (symbol->string type) "-"))))
       ; an id string
       (id-str (symbol->string id-sym))
       ; determine pointer for the parent
       (parent-ptr (if parent-id
                     (send hierarchy-widget item-property-get parent-id 'pointer #f)
                     #f))
       ; generate a preview
       (ptr (preview-create-widget id-str item-props parent-ptr))
      )
      ; assemble proerty-list
      (append (list id-str (list 'pointer ptr))
              item-props
      )
    )
  )
  
  ;; This function creates a MrEd widget for preview
  (provide preview-create-widget)
  (define (preview-create-widget id line parent)
    (let 
      ((type (cadr (assoc 'type line)))
       (line line)
      )
      (cond
        ((eq? type 'frame)                (frame-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'button)               (button-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'radio-box)            (radio-box-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'check-box)            (check-box-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'pane)                 (pane-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'horizontal-pane)      (horizontal-pane-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'vertical-pane)        (vertical-pane-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'panel)                (panel-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'horizontal-panel)     (horizontal-panel-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'vertical-panel)       (vertical-panel-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'tab-panel)            (tab-panel-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'group-box-panel)      (group-box-panel-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'grow-box-spacer-pane) (grow-box-spacer-pane-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'message)              (message-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'slider)               (slider-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'gauge)                (gauge-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'list-box)             (list-box-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'choice)               (choice-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'text-field)           (text-field-default-preview 
                                             id line parent preview-clicked-callback))
        ((eq? type 'canvas)               (canvas-default-preview 
                                             id line parent preview-clicked-callback))
        (else
          (error "Unknown type: " type)
        )
      )
    )
  )
  
  ; this function shows or hides all frames in the hierarchy list
  (provide preview-show-frames)
  (define (preview-show-frames line show?)
    (cond
      ((null? line) #f)
      ((list? (car line))
       (preview-show-frames (car line) show?)
       (preview-show-frames (cdr line) show?))
      (else
       (let*
         ((elem (car line))
          (type (send hierarchy-widget item-property-get elem 'type #f)))
         (if (eq? type 'frame)
           (let ((window (send hierarchy-widget item-property-get elem 'pointer #f)))
             (if show?
               (if (not (send window is-shown?))
                 (send window show #t))
               (if (send window is-shown?)
                 (send window show #f))))))
       (preview-show-frames (cdr line) show?))))
  
  ; this function goes through the list and determines whether the id is in the list
  ; we have to consider atoms and cars of lists
  (define (find-children id lst)
    (cond
      ((null? lst) #f)
      ((and (list? (car lst)) (equal? (caar lst) id)) #t)
      ((and (not (list? (car lst))) (equal? (car lst) id)) #t)
      (else (find-children id (cdr lst)))
    )
  )
  
  ; it goes through a list and determines where the id is,
  ; the function also determines all neighbouring children and the parent container
  ; It returns #f or a list where the car is the parent and the cdr contains
  ; the ids of all neighbouring children 
  (define (find-children-list id lst)
    (cond
      ((null? lst) #f)
      ((list? (car lst))
       (if (find-children id (cdar lst))
         ; if we have found the id then
         ; return the container and the list of children in one list
         (cons (caar lst) (map (lambda (x) (if (list? x) (car x) x)) (cdar lst)))
         (let ((res (find-children-list id (car lst))))
           (if res
             res
             (find-children-list id (cdr lst))
           )
         )
       )
      )
      (else
        (find-children-list id (cdr lst))
      )
    )
  )
  
  ; the function removes from the id list all of those ids that correspond to a frame
  (define (remove-children-frame lst)
    (cond
      ((null? lst) '())
      ((equal? (send hierarchy-widget item-property-get (car lst) 'type #f) 'frame)
       (remove-children-frame (cdr lst))
      )
      (else
        (cons (car lst) (remove-children-frame (cdr lst)))
      )
    )
  )

;; This function updates, in the preview, the order of the element's children
;; This function is valid for swap-up and swap-down
(provide preview-update-children-order)
(define (preview-update-children-order id)  
  (let*
    ((lst (list (send hierarchy-widget get-hierarchy)))
     (children (find-children-list id lst)))
    ; if there are children and the list is not empty
    (if (and children (not (null? children)))
      ; project is not updated after a swap
      (if (not (equal? (car children) "project"))
        ;; We ask the widget to update its children
        ;; any child that is a frame is left out of this update
        (let
          ((ptr (send hierarchy-widget item-property-get (car children) 'pointer #f)))
          (send ptr change-children 
                (lambda (lst)
                  (map (lambda (x) (send hierarchy-widget item-property-get x 'pointer #f))
                       (remove-children-frame (cdr children))
                  )
                )
          )
        )
      )
      ; empty list is a problem, signal error
      (begin
        (error (string-append "ID: " id " is not found in hierarchy: ") lst)
      )
    )
  )
)


;; This function is called for example when the user opens a new file : it
;; then removes the preview from the previous file and creates the preview
;; from the new file...
;; It is also called when the user updates the preview "on the fly" (when
;; the user changes the style property of a container).
(provide preview-update-all)
(define (preview-update-all)
  (define (create-widget lst parent)
    (cond
      ((null? lst) #f)
      ((list? (car lst))
       (let
         ((ptr (preview-create-widget (caar lst) 
                                      (send hierarchy-widget get-item (caar lst))
                                      parent)))
         (send hierarchy-widget item-property-set! (caar lst) 'pointer ptr #f)
         (create-widget (cdar lst) ptr)
       )
       (create-widget (cdr lst) parent)
      )
      (else
        (let
          ((ptr (preview-create-widget (car lst) 
                                       (send hierarchy-widget get-item (car lst))
                                       parent)))
          (send hierarchy-widget item-property-set! (car lst) 'pointer ptr #f)
        )
        (create-widget (cdr lst) parent)
      )
    )
  )
  (send hierarchy-widget update-view! #f)
  (let
    ((lst (send hierarchy-widget get-hierarchy)))
    ;; create the frames (windows)
    (create-widget (cdr lst) #f)
  )
  (send hierarchy-widget update-view! #t)
  (send hierarchy-frame focus)
)

  
  
(provide preview-update-widget)
(define (preview-update-widget id prop val)
  ;; If the widget is a container and the last value changed is a property that we
  ;; cannot change "on the fly", then, we build the whole thing
  ; FIXME: alternatively we can build a new widget with the 
  ; FIXME:   property and replace with change-children
  (if (or (eq? 'style prop)
          (eq? 'min-value prop)
          (eq? 'max-value prop)
          ; in this way widgets can be created without labels
          (eq? 'label prop)
          (eq? 'choices prop)
;          (eq? 'callback prop)
      )
    (preview-update-all)
    ;; Otherwise, we will just change each property separately.
    (begin
      (let ((widget-pointer (send hierarchy-widget item-property-get id 'pointer #f)))
        ;;
        ;; And then we have (almost) no choice, because MrEd's function names differ,
        ;; and there is no real rule, therefore, we have to enumerate each case!
        ;; (We could, of course, make the code a bit shorter but it wouldn't be
        ;; worth doing it).
        ;;
        (case prop
          ;; These ones begin with "set"...
          ((label) 
           (send widget-pointer set-label val))
          ((selection) 
           (send widget-pointer set-selection val))
          ((value) 
           (send widget-pointer set-value val))
          ((init-value) 
           (send widget-pointer set-value val))
          ((range) 
           (send widget-pointer set-range val))
              
          ;; These ones have no prefix...
          ((horiz-margin) 
           (send widget-pointer horiz-margin val))
          ((vert-margin) 
           (send widget-pointer vert-margin val))
          ((min-width) 
           (send widget-pointer min-width val))
          ((min-height) 
           (send widget-pointer min-height val))
          ((stretchable-width) 
           (send widget-pointer stretchable-width val))
          ((stretchable-height) 
           (send widget-pointer stretchable-height val))
          ((border)
           (send widget-pointer border val))
          ((spacing)
           (send widget-pointer spacing val))
              
          ;; These ones are really weird !
          ((width)
           (send widget-pointer resize val (send widget-pointer get-height)))
          ((height) 
           (send widget-pointer resize (send widget-pointer get-width) val))
          ((x) 
           (send widget-pointer move val (send widget-pointer get-y)))
          ((y)
           (send widget-pointer move (send widget-pointer get-x) val))
          ((enabled) 
           (send widget-pointer enable val))
          ((alignment) 
           (send widget-pointer set-alignment (car val) (cadr val))
          )
        ) ; end of case
      ) ; end of let
    ) ; end of begin
  ) ; end of if
)

) ; end of module



