;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - widgets.ss                                                     #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # Copyright (C) Peter Ivanyi, 2007                                               #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

(module mreddesigner-widgets mzscheme
  (require (lib "class.ss")  
           (lib "mred.ss" "mred")
           (file "mreddesigner-hierarchy-widget.ss")
           (file "mreddesigner-properties-widget.ss")
  )
  
  (provide hierarchy-frame
           hierarchy-up-button
           hierarchy-up-one-button
           hierarchy-down-one-button
           hierarchy-down-button
           hierarchy-delete-button
           hierarchy-show-button
           hierarchy-hide-button
           hierarchy-widget
  )
  (define hierarchy-frame #f)
  (define hierarchy-up-button #f)
  (define hierarchy-up-one-button #f)
  (define hierarchy-down-one-button #f)
  (define hierarchy-down-button #f)
  (define hierarchy-delete-button #f)
  (define hierarchy-show-button #f)
  (define hierarchy-hide-button #f)
  (define hierarchy-widget #f)

  ; this function initialises the window with the hierarchy frame
  (provide hierarchy-frame-init)
  (define (hierarchy-frame-init on-close-func
                                hierarchy-up-func
                                hierarchy-up-one-func
                                hierarchy-down-one-func
                                hierarchy-down-func
                                hierarchy-delete-func
                                hierarchy-hide-func
                                hierarchy-show-func
                                hierarchy-widget-func
                                hierarchy-widget-key-func
                                )
    ;; We need to derive the frame class to override the on-close method.
    (define hierarchy-frame%
      (class frame%
        ;; When on-close is called, we exit the project.
      (define/augment (on-close) (on-close-func))
      (super-instantiate ())))
    
    ; create the hierarchy frame
    (set! hierarchy-frame (new hierarchy-frame%
                               (label "Hierarchy")
                               (parent #f)
                               (x 5)
                               (y 190)
                               (width 300)
                               ;(width (send toolbox-frame get-width))
                          ))
    (let*
      ((hierarchy-horizontal-panel (new horizontal-panel%
                                        (border 0)
                                        (parent hierarchy-frame)))
       ;; ...the buttons' panel...
       (buttons-panel (new vertical-panel%
                           (border 1)
                           (alignment '(center top))
                           (stretchable-width #f)
                           (parent hierarchy-horizontal-panel))))
      ;; ...the swap up button...
      (set! hierarchy-up-button
            (new button% 
                 (label (make-object bitmap%
                                     (build-path "images" "hierarchy-up.png")
                                     'png (make-object color% 255 255 255)))
                 (callback hierarchy-up-func)
                 (enabled #f)
                 (parent buttons-panel)))
      
      ;; ...the swap up one button...
      (set! hierarchy-up-one-button 
            (new button% 
                 (label (make-object bitmap% 
                                     (build-path "images" "hierarchy-up-one.png")
                                     'png (make-object color% 255 255 255)))
                 (callback hierarchy-up-one-func)
                 (enabled #f)
                 (parent buttons-panel)))
      
      ;; ...the swap down one button...
      (set! hierarchy-down-one-button 
            (new button% 
                 (label (make-object bitmap% 
                                     (build-path "images" "hierarchy-down-one.png")
                                     'png (make-object color% 255 255 255)))
                 (callback hierarchy-down-one-func)
                 (enabled #f)
                 (parent buttons-panel)))
      
      ;; ...the swap down button...
      (set! hierarchy-down-button 
            (new button% 
                 (label (make-object bitmap% 
                                     (build-path "images" "hierarchy-down.png")
                                     'png (make-object color% 255 255 255)))
                 (callback hierarchy-down-func)
                 (enabled #f)
                 (parent buttons-panel)))
      
      ;; ...the delete button...
      (set! hierarchy-delete-button 
            (new button% 
                 (label (make-object bitmap% 
                                     (build-path "images" "hierarchy-delete.png")
                                     'png (make-object color% 255 255 255)))
                 (callback hierarchy-delete-func)
                 (enabled #f)
                 (parent buttons-panel)))
      
      ;; ...the hide button...
      (set! hierarchy-hide-button 
            (new button% 
                 (label (make-object bitmap% 
                                     (build-path "images" "hierarchy-hide.png")
                                     'png (make-object color% 255 255 255)))
                 (callback hierarchy-hide-func)
                 (enabled #f)
                 (parent buttons-panel)))

      ;; ...the show button...
      (set! hierarchy-show-button 
            (new button% 
                 (label (make-object bitmap% 
                                     (build-path "images" "hierarchy-show.png")
                                     'png (make-object color% 255 255 255)))
                 (callback hierarchy-show-func)
                 (enabled #f)
                 (parent buttons-panel)))
      
      ; the hierarchy widget itself
      (set! hierarchy-widget (new hierarchy-widget% 
                                  (parent hierarchy-horizontal-panel)
                                  (callback hierarchy-widget-func)
                                  (keyboard-callback hierarchy-widget-key-func)
                                  ))
    )
  )
  
  
  ;;; --------------------------------------------------------------------------
  ;;; toolbox-frame
  ;;; --------------------------------------------------------------------------
  
  ;; Toolbox frame definition...
  (provide toolbox-frame
           toolbox-frame-init
  )
  (define toolbox-frame #f)
  (define (toolbox-frame-init on-close-func)
    (define toolbox-frame%
      (class frame%
        ;; When on-close is called, we exit the project.
        (define/augment (on-close) (on-close-func))
        (super-instantiate ())))
    (set! toolbox-frame (new toolbox-frame%
                             (label "Toolbox")
                             (parent #f)
                             (width 355)
                             (x 5)
                             (y 5)
                             (style '(no-resize-border)))))

  ;;; --------------------------------------------------------------------------
  ;;; properties-frame
  ;;; --------------------------------------------------------------------------
  
  (provide properties-frame
           properties-panel
           properties-frame-init
  )
  (define properties-frame #f)
  (define properties-panel #f)
  (define (properties-frame-init on-close-func action-func)
    (define properties-frame%
      (class frame%
        ;; When on-close is called, we exit the project.
        (define/augment (on-close) (on-close-func))
        (super-instantiate ())))
    
    (let-values (((screen-w screen-h) (get-display-size)))
      (set! properties-frame (new properties-frame%
                                  (label "Properties")
                                  (parent #f)
                                  (x (- screen-w 310))
                                  (y 5)
                                  (min-width 300)
                                  (min-height 100)
                                  (stretchable-height #f)
;                                  (style '(no-resize-border))
                                  )))
    (set! properties-panel (new property-panel% 
                                (parent properties-frame)
                                (callback action-func)))
    (send properties-panel empty)
  )
  
  
) ; end of module

